%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Pradeep Nagesh, Arizona State University, USA (2007-2008)  
%         pnagesh@asu.edu
%         Face Recognition using Joint Sparsity Models
%
% Description: Perform training - Extract the Common and Innovations and
% form the sum of Innovations from mean subtracted training images and store 
% the results as .mat files and also as images for inspection. Use the
% pre-generated JSM data for training.
%  
%  Input/Output 
%  Options: Input data structure containing various parameter settings.  
%  Refer to FR_Document.doc for details.
%
%  Face_classOmega: The current Class number on which feature extraction is
%  to performed. Face_classOmega = 1...L
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%        Face Recognition using JSM -1 Models                           %%

function Status = FaceDatabase_Train_Class(Options, Face_classOmega)

NumImages = Options.M(Face_classOmega);
Options.ClassNum = Face_classOmega;

% image length and width
N1 = Options.SizeImage(1);
N2 = Options.SizeImage(2);
SampleFraction = Options.SampleFraction;

% Seperate 1D and 2D feature extraction. 1D simply means that the images are
% vectored and operated (see formulation in "A Compressive Sensing Approach for 
% Expression Invarient Face Recognition", CVPR). Usually 2D is preferred because 
% of faster computation and lower memory requirement. 1D feature extraction
% is often not feasible in MATLAB if the number of images per class and the
% image sizes are large due to "OUT of Memory" problems.

if (0 == Options.twoDReconstruction)
    N =N1*N2;
    idctmx = khronical_prdct(dctmtx(N1)');
    % number of observations to make
    K = round(SampleFraction*N1*N2);
else
    N = N1;
    K = N;
    idctmx = zeros(N1,N2);
    idctmx = dctmtx(N)';
end

% Get the appropriate stacked mean subtracted measurements for each class   
% y_ms and tet = measurement_matrix*basis_matrix which are already stored 
% for BJSM Case
% ms --> Stands for Mean Subtracted in all the variables user below.
[y_ms, tet]= Load_JSMData(Options, Face_classOmega);
tet_inv = tet';
% Form initial estimate of sparse solution of mean subtracted measurements
x0_ms = tet_inv*y_ms;
clear tet_inv;

% for SJSM case
%[y_SJSM_ms, tetSJSM]= Load_JSMDataSJSM(Options, Face_classOmega);
%tet_inv = tetSJSM';
%x0_SJSM_ms = tet_inv*y_SJSM_ms;
%clear tet_inv;

if (0 == Options.twoDReconstruction)
    % If using l1 magic package -- note this solves constrained Basis
    % Pursuit problem for 1-D case only -- very slow, not prefered
    if (strcmp(Options.Optimization_Algorithm,'l1magic'))
        max_Itr = 28;
        err = 1e-3;
        for steps=1:4
            xp = l1eq_pd(x0,tet,[], y,err,max_Itr);
            x0 = xp;
        end
    % If using GPSR package -- note this solves unconstrained l1 minimization 
    % (Basis Pursuit Denoising) problem for 1-D case.
    elseif (strcmp(Options.Optimization_Algorithm,'GPSR'))

        tau = 0.009;
        first_tau_factor = 0.8*(max(abs(x0))/tau);
        steps = 20;
        debias = 0;
        stopCri=3;
        tolA=1.e-10;

        for i=1:1    
            xp = GPSR_BB(y,tet,tau,...
                 'AT', [],...
                 'Debias',0,...
                 'Continuation',1,...
                 'ContinuationSteps',steps,...
                 'FirstTauFactor',first_tau_factor,...
                 'Monotone',1,...
                 'Initialization',x0,...
                 'StopCriterion',stopCri,...
                 'ToleranceA',tolA,...
                 'Verbose',1);
             x0 = xp;
        end
    end

else
    % Come here if l1 minimization is done on 2D signals  
    % Use the GPSR set-up below - other algorithms like TV minimization can
    % also be used.
    if (strcmp(Options.Optimization_Algorithm,'GPSR'))    
        tau = 0.009;
        x0 = tet'*y_ms;
        first_tau_factor = 0.8*(max (max((abs(x0)/tau))));
        steps = 5;
        debias = 0;
        stopCri=3;
        tolA=1.e-8;

        % This is for BJSM. In case SJSM is required, simply replace xp_ms,
        % y_ms and x0_ms by xp_SJSM_ms, y_SJSM_ms, x0_SJSM_ms.
        for i= 1:1
            xp_ms = GPSR_BB(y_ms,tet,tau,...
                 'Debias',0,...
                 'Continuation',1,...
                 'ContinuationSteps',steps,...
                 'FirstTauFactor',first_tau_factor,...
                 'Monotone',1,...
                 'Initialization',x0_ms,...
                 'StopCriterion',3,...
                 'ToleranceA',tolA,...
                 'Verbose',0);
             x0_ms = xp_ms;
        end
    end
end    

%%   Seperate the spatial domain innovation and common components %%

clear tet;
clear y;
x_out = [];

% If solving in 1D, using more than 5 images per class in training is not
% supported -- this is because of the likely Out of Memory problems that
% can occur due to huge sizes of basis,measurement matrix product.

if (0 == Options.twoDReconstruction)
    if(NumImages > 1)
        sc = xp(1:N);
        sp_1 = xp((N+1):(2*N));
        sp_2 = xp((2*N+1):(3*N));

        x_c = idctmx * sc;
        xp_1 = idctmx * sp_1;
        xp_2 = idctmx * sp_2;

        x_c = reshape(x_c,N1,N2);
        xp_1 = reshape(xp_1,N1,N2);
        xp_2 = reshape(xp_2,N1,N2);
        x_inn = [xp_1; xp_2];
    end
    if(NumImages > 2)
        sp_3 = xp((3*N+1):(4*N));
        xp_3 = idctmx * sp_3;
        xp_3 = reshape(xp_3,N1,N2);
        x_inn = [x_inn; xp_3];
    end
    if(NumImages > 3)
        sp_4 = xp((4*N+1):(5*N));
        xp_4 = idctmx * sp_4;
        xp_4 = reshape(xp_4,N1,N2);
        x_inn = [x_inn; xp_4];
    end
    if(NumImages > 4)
        sp_5 = xp((5*N+1):(6*N));
        xp_5 = idctmx * sp_5;
        xp_5 = reshape(xp_5,N1,N2);
        x_inn = [x_inn; xp_5];
    end
    if(NumImages > 5)
        sp_6 = xp((6*N+1):(7*N));
        xp_6 = idctmx * sp_6;
        xp_6 = reshape(xp_6,N1,N2);
        x_inn = [x_inn; xp_6];
    end
else

% Sepeate the common and innovations and wtite them into mat files and as
% images for the 2D case.
    Seperate_the_Components(xp_ms, Options, Face_classOmega, idctmx, 'BJSM_MS');
    %I =eye(N)
    %Seperate_the_Components(xp_ms_SJSM, Options, Face_classOmega, I, 'SJSM_MS');
end

% Set every thing went well, status is complete
Status = 1;

%%                       Function to Load JSM Data                       %%                          

function [y_ms tet] = Load_JSMData(Options, ClassNum)

Opts = Options;
Opts.ClassNum = ClassNum;
load(GetPath_name('JSM Data', Opts),'y_ms','tet');

function [y_SJSM y_ms_SJSM tet_SJSM] = Load_JSMData_SJSM(Options, ClassNum)

Opts = Options;
Opts.ClassNum = ClassNum;
load(GetPath_name('JSM Data', Opts),'y_ms_SJSM' ,'tet_SJSM');

%% Function to Seperate the 2D JSM training results

function Seperate_the_Components(xp, Options, ClassNum, Matrix, Operation)

temp = Options.M;
NumImages = temp(ClassNum)
N = size(xp,1)./(NumImages+1);
N1 = Options.SizeImage(1);
N2 = Options.SizeImage(2);
sc = xp(1:N,1:N);
x_inn_all = 0;
x_c = Matrix*sc;
x_inn = zeros(N1,N2*NumImages);

for i=2:NumImages+1
    temp_sp = xp((i-1)*N+1:i*N,1:N);    
    temp_xp = Matrix*temp_sp;    
    x_inn(1:N1,(i-2)*N2+1:(i-1)*N2) =  temp_xp;
    x_inn_all = x_inn_all + temp_xp;
end

WriteImages(x_c, x_inn, x_inn_all, Options,ClassNum, Operation);

%%                     Function to Write Results                         %%
function WriteImages(x_c, x_inn, x_inn_all, Options,ClassNum, Operation)

Opts = Options;
Opts.ClassNum = ClassNum;

if (strcmp(Operation, 'BJSM'))
    [nameout_image nameout_image1] = GetPath_name('Training Images',Opts);
    save(GetPath_name('Training Data',Opts),'x_c','x_inn', 'x_inn_all');
elseif (strcmp(Operation, 'BJSM_MS'))
    [nameout_image nameout_image1] = GetPath_name('Training Images MS',Opts);
    save(GetPath_name('Training Data MS',Opts),'x_c','x_inn', 'x_inn_all');
elseif (strcmp(Operation, 'SJSM'))
    [nameout_image nameout_image1] = GetPath_name('Training Images SJSM',Opts);
    save(GetPath_name('Training Data SJSM',Opts),'x_c','x_inn', 'x_inn_all');
elseif (strcmp(Operation, 'SJSM_MS'))    
    [nameout_image nameout_image1] = GetPath_name('Training Images MS SJSM',Opts);
    save(GetPath_name('Training Data MS SJSM',Opts),'x_c','x_inn', 'x_inn_all');
end

imwrite(x_c, nameout_image, 'jpeg');
imwrite([x_inn x_inn_all], nameout_image1, 'jpeg');