%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Pradeep Nagesh, Arizona State University, USA (2007-2008)  
%         pnagesh@asu.edu
%         Face Recognition using Joint Sparsity Models
%
% Description: Perform testing - Implement the classifier using JSM models.
%
%  Input/Output 
%  Options: Input data structure containing various parameter settings.  
%  Refer to FR_Document.doc for details.
%
%  testImage_name: The current test image name in the folder
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function class_res1 = FaceTestClass_Sparsity_Distance(testImage_name, Options)

tempNumImages = Options.M;
NumClasses = Options.L;
SampleFraction = Options.SampleFraction;

% Sample Fraction or %age measurements for Low-Dimensional Recognition
FRAC =  SampleFraction/100;
class_res1 = 0;

% Read the test image
x = double(imread(testImage_name))./255;

% image sizes
N1 = Options.SizeImage(1);
N2 = Options.SizeImage(2);

if ((N1~= size(x,1)) || (N1~= size(x,2)))
    error('Specified Image training feature size and test image size does not match');
end

% Get the stored data - like measurement matrix, basis matrix etc
Options.ClassNum =1; % dummy entry
load(GetPath_name('JSM Test Data',Options),'Phi','tet','idctmx');
tet_inv = tet';
I = eye(N1);

if (0 == Options.twoDReconstruction)
    N = N1*N2;
    K = round((Options.SampleFraction/100)*N);       
else
    N = N1;
end

global ClassRes;
global test_img_counter;

x_test = x;
L = Options.L;
s2_MS = zeros(1,L);

% Setting Save Every Result saves the intermediate results for a test image
% with all class trainig features. i.e the images reconstructed using training
% features of all classes and the final residual.
if (Options.SaveEveryResult)
    z_est_MS_FULL = zeros(N1,N2*L); 
    d_c_MS_FULL = zeros(N1,N2*L);
end

sprintf('Testing.....\nCurrent Database: %s',Options.DataBaseName);
sprintf('Current Image: %s',testImage_name(end-9:end));

% Only 2D testing is implemented. It can be changed to 1D implementation if
% the test image and training features are vectored and if the other matrices
% involved are appropiaty derived (dimensions of Phi and Basis Matrix)

for class=1:NumClasses   
    Opts = Options;
    Opts.ClassNum = class;
    NumImages = tempNumImages(class);
    
    % GPSR Algorithm: Implemnted for 100% measurement - i.e Classification 
    % is not done in Low dimensional features.
    if (strcmp(Options.Optimization_Algorithm,'GPSR'))

        sprintf('class = %d',class)
        x_BJSM_MS_inp = zeros(N1*2,N2);
        x_BJSM_MS_ginn = zeros(N1*2,N2);

        TrainingDataPath = GetPath_name('Training Data MS',Opts);
        load (TrainingDataPath, 'x_c', 'x_inn', 'x_inn_all');
        mean_xs = mean(mean(x_test));
        x_test_ms = x_test - mean_xs;
        x_d = x_test_ms - x_c;
        x_BJSM_MS_inp = [x_test_ms; (x_c+x_inn_all)];
        x_BJSM_MS_ginn = [x_d; x_inn_all];

        y_ms_c = Phi*x_BJSM_MS_inp;
        y_ms_gi = Phi*x_BJSM_MS_ginn;
        xp_ms_c = JSM_Test_Optimization_GPSR(y_ms_c,tet);            
        xp_ms_gi = JSM_Test_Optimization_GPSR(y_ms_gi,tet);            
        [s z_est d_c] = Classification_Data_New(x_test_ms, x_c,xp_ms_gi,idctmx);
        
        s2_MS(1,class) = s(1);  % Change Index on Uncomenting to 3

        if (Options.SaveEveryResult)
            z_est_MS_FULL(1:N1,(class-1)*N2+1:class*N2) = z_est;
            d_c_MS_FULL(1:N1,(class-1)*N2+1:class*N2) = d_c;
        end
        
    % TV Algorithm: Implemnted for any measurement - i.e Classification 
    % can be done in Low dimensional features by setting appropriate sample
    % fraction FRAC.
    elseif (strcmp(Options.Optimization_Algorithm,'TV_L1'))
       
        sprintf('class = %d',class)
        
        Gmask1=zeros(N1,N2);
        ind=randperm(N1*N2);
        Gmask1(ind(1:round(N1*N2*FRAC)))=1;

        Gmask2=zeros(N1,N2);
        ind=randperm(N1*N2);
        Gmask2(ind(1:round(N1*N2*FRAC)))=1;

        Gmask  = [Gmask1; Gmask2;];
        mask=Gmask;
        
        load(GetPath_name('JSM Test Data',Options),'Phi','shi_tilda','idctmx');
        x_BJSM_MS_ginn = zeros(N1*2,N2);
       
        TrainingDataPath = GetPath_name('Training Data MS',Opts);
        load (TrainingDataPath, 'x_c', 'x_inn', 'x_inn_all');
        mean_xs = mean(mean(x_test));
        x_test_ms = x_test - mean_xs;
        x_d = x_test_ms - x_c;
        x_BJSM_MS_ginn = [x_d; x_inn_all];

        y_ms_gi = Phi*x_BJSM_MS_ginn;
        
        %generate sampling operator
        FT = matrixproj(mask, size(y_ms_gi), Phi);
        
        % scale data
        im_dc = FT'*(y_ms_gi.*mask);
        im_dc = im_dc/max(abs(im_dc(:)));       
 
        TVWeight = 4e-3; 	% Weight for TV penalty
        xfmWeight = 4e-3;	% Weight for Transform L1 penalty
        Itnlim = 8;		    % Number of iterations
        param = init;
        param.FT = FT;
        param.XFM = shi_tilda';
        param.TV = TVOP;
        param.data = y_ms_gi;
        param.TVWeight =TVWeight;     % TV penalty 
        param.xfmWeight = xfmWeight;  % L1 wavelet penalty
        param.Itnlim = Itnlim;
        
        res = shi_tilda'*im_dc;

        % do iterations
        for n=1:8
            res = fnlCg(res,param);
        end
        xp_ms_gi = res;
        
        [s z_est d_c] = Classification_Data_New(x_test_ms, x_c,xp_ms_gi,idctmx);
        
        s2_MS(1,class) = s(1);  % Change Index on Uncomenting to 3

        if (Options.SaveEveryResult)
            z_est_MS_FULL(1:N1,(class-1)*N2+1:class*N2) = z_est;
            d_c_MS_FULL(1:N1,(class-1)*N2+1:class*N2) = d_c;
        end
    end
end

%% Done with solving the l1 minimization problem: Now simply wind-up

% Reconize to which of the databases (CK, JAFFE, CMU) does the test image
% belong to and accordingly determine the actual class to which it belongs
% to from the naming convention. If new database is to be added, then the
% actualClass need to be determined by decoding the appropriate naming
% convention for the class.

TestResSavepath = strcat(testImage_name(1:end-4),'.mat');
[minVal class_MS_res2] = min(s2_MS);

if ~isempty(strfind(Options.DataBaseName,'DataBaseName1'))
    actualClass = @(z) ( (str2num(sprintf('%d',z(1))) - 64));
end

if ~isempty(strfind(Options.DataBaseName,'Expression'))
    actualClass = @(z) ( (str2num(sprintf('%d',z(1))) - 64));
end

if ~isempty(strfind(Options.DataBaseName,'CK'))
    actualClass = @(z) ( str2num(z(1:2)) );
end

if ~isempty(strfind(Options.DataBaseName,'JAFFE'))
    actualClass = @(z) JAFFE_Class(z);
end

if ~isempty(strfind(Options.DataBaseName,'DataBaseName1'))
    testFname = @(z) (z(end-6:end));
end

if ~isempty(strfind(Options.DataBaseName,'Expression'))
    testFname = @(z) (z(end-6:end));
end

if ~isempty(strfind(Options.DataBaseName,'JAFFE'))
    testFname = @(z) z(end-9:end);
end

if ~isempty(strfind(Options.DataBaseName,'CK'))
    testFname = @(z) z(end-23:end);
end

TestResFull_SavePath = strcat(Options.DataBaseName,'.mat');

% Save Every Result: Saves even the images and residuals; this option can 
% be invoked for debugging purposes, but otherwise will be lot of data!!
if (Options.SaveEveryResult)
    save(TestResSavepath,'z_est_MS_FULL', 'd_c_MS_FULL',...
                         'class_MS_res2',...
                         's2_MS');
    TestData_RootPath = GetPath_name('Testing Data',Options);
    imwrite(z_est_MS_FULL,strcat(TestResSavepath,'_z_MS_est','.bmp'));                     
end


resAll_MS=[];
actCass = actualClass(testFname(testImage_name));
resAll_MS = [actCass; class_MS_res2];

%  Global Result: This is only clasification results: No intermediate
%  results like estimated images or residuals are saved.
ResGlobal_Savepath = strcat((GetPath_name('Global Result Save',Options)),'.mat');

% Done with everything, simply save the actual class and the classification
% result on to ClassRes and save it. For next image in the same folder
% simply append the next result to ClassRes and save again.
try
    load(ResGlobal_Savepath);
    ClassRes(:,test_img_counter) = resAll_MS;
    save(ResGlobal_Savepath, 'ClassRes');
catch
    ClassRes(:,test_img_counter) = resAll_MS;
    save(ResGlobal_Savepath, 'ClassRes');
end

%% JSM Model l1 Minimization

function xp = JSM_Test_Optimization_GPSR(y,tet)

x0 = tet'*y;

tau = 0.009;
first_tau_factor = 0.8*(max (max((abs(x0)/tau))));
steps = 5;

debias = 0;

stopCri=3;
tolA=1.e-8;

for k= 1:1

    xp = GPSR_BB(y,tet,tau,...
         'Debias',0,...
         'Continuation',1,...
         'ContinuationSteps',steps,...
         'FirstTauFactor',first_tau_factor,...
         'Monotone',1,...
         'Initialization',x0,...
         'StopCriterion',3,...
         'ToleranceA',tolA,...
         'Verbose',0);
     x0 = xp;
end

%% Classification Algorithms

%All tried Classification Algorithms including sparsities
function [s z_est x_c_c d_c] = Classification_Data_Old(x_test,x_c,x_ca, xp1,xp2,Matrix, thr)
    
[N1 N2] = size(Matrix);

[x_c_c z_inn x_c_inn] = Common_Innvation2D(xp1,Matrix);
[d_c d_inn x_inn_all_inn] = Common_Innvation2D(xp2,Matrix); 

K_C = Sparsity_Calculate(xp1(1:N1,1:N2),thr);
K_Z_INN = Sparsity_Calculate(xp1(N1+1:2*N1,1:N2),thr);
K_XC_INN = Sparsity_Calculate(xp1(2*N1+1:3*N1,1:N2),thr);
K_Total1 = K_C + K_Z_INN + K_XC_INN;

K_X_CA = Sparsity_Calculate(x_ca,thr);
K_Z = Sparsity_Calculate(x_test,thr);
K_Total = K_Z + K_X_CA;

d0_0 = (z_inn.*z_inn);
temp_s0 = sum(sum(d0_0));

temp_s1 = (K_Total1/K_Total);
z_est = x_c + d_c;
d2 = x_test - z_est;
temp_s2 = sum(sum(d2.*d2));

s = [temp_s0 temp_s1 temp_s2];

% Only the New Classification as used in "A Compressive Sensing Approach for 
% Expression Invarient Face Recognition"

function [s z_est d_c] = Classification_Data_New(x_test, x_c, xp2,Matrix)
    
[N1 N2] = size(Matrix);
[d_c d_inn x_inn_all_inn] = Common_Innvation2D(xp2,Matrix); 

z_est = x_c + d_c;
d2 = x_test - z_est;
temp_s2 = sum(sum(d2.*d2));

s = temp_s2;

%%   Seperate the in spatial domain innovation and common components %%

clear tet;
clear y;

function [x_c x_inn1 x_inn2] = Common_Innvation(xp, imatrix, size)

N_3 = length(xp);
N = N_3/3;

N1 = size(1);
N2 = size(2);

sc = xp(1:N);
sp_1 = xp((N+1):(2*N));
sp_2 = xp((2*N+1):(3*N));
    
x_c = imatrix * sc;
xp_1 = imatrix * sp_1;
xp_2 = imatrix * sp_2;
    
x_inn1 = xp_1(:);
x_inn2 = xp_2(:);

function [x_c x_inn1 x_inn2] = Common_Innvation2D(xp, imatrix)

N_3 = size(xp,1);
N2 = size(xp,2);
N1 = N_3/3;

sc = xp(1:N1,1:N2);
sp_1 = xp((N1+1):(2*N1),1:N2);
sp_2 = xp((2*N1+1):(3*N1),1:N2);
    
x_c = imatrix * sc;
x_inn1 = imatrix * sp_1;
x_inn2 = imatrix * sp_2;

%%                     Function to Write Results                         %%

function WriteImages1(x_c, x_inn, N, NumImages,SampleFraction,ClassNum)

Opts.SampleFraction = SampleFraction;
Opts.ClassNum = ClassNum;
Opts.NumImages = NumImages;

nameout_image = GetPath_name('Test Images',Opts);
imwrite(x_c, nameout_image, 'jpeg');

for i=1:NumImages
    Opts.NumImages = i;

    [name_dummy nameout_image] = GetPath_name('Training Images',Opts);
    imwrite(x_inn(1+((i-1)*N):(N*i),:), nameout_image, 'jpeg');
end

save(GetPath_name('Training Data',Opts),'x_c','x_inn');

%%  Decode JAFFE database class

function act_class = JAFFE_Class(z)

act_class = str2num(z(3));

if (strcmp(z(2:3), '10'))
    act_class = 10;
end